from decimal import Decimal
from warnings import warn

LOVELACE = Decimal("0.000001")


def to_lovelaces(amount):
    """
    Convert ADA to Lovelaces.

    :param Decimal,int amount:    the amount of ADA
    :rtype: :class:`int`

    """
    if not isinstance(amount, (Decimal, int, float)):
        raise ValueError(
            "Amount '{}' doesn't have numeric type. Only Decimal, int and "
            "float (not recommended) are accepted as amounts."
        )
    if isinstance(amount, float):
        warn(
            "to_lovelaces() received amount of float type ({:f}). It is STRONGLY DISCOURAGED "
            "to use floating-point numbers for monetary operations due to rounding errors".format(
                amount
            ),
            RuntimeWarning,
        )
    return int(amount * 10 ** 6)


def from_lovelaces(amount):
    """
    Convert Lovelaces to ADA.

    :param int amount:  the amount of Lovelaces
    :rtype: :class:`Decimal`
    """
    return (Decimal(amount) * LOVELACE).quantize(LOVELACE)


def as_ada(amount):
    """
    Return the amount rounded to maximal ADA precision.

    :param Decimal,int amount:  the amount to be sanitized
    :rtype: :class:`Decimal` with 6 decimal places precision
    """
    if isinstance(amount, float):
        warn(
            "as_ada() received amount of float type ({:f}). It is STRONGLY DISCOURAGED "
            "to use floating-point numbers for monetary operations due to rounding errors".format(
                amount
            ),
            RuntimeWarning,
        )
    return Decimal(amount).quantize(LOVELACE)



from ... import exceptions


class WalletRESTException(exceptions.BackendException):
    def __init__(self, *args, **kwargs):
        self.result = kwargs.pop("result", None)
        super(WalletRESTException, self).__init__(*args, **kwargs)


class BadRequest(WalletRESTException):
    pass


class NotFound(WalletRESTException):
    pass



from .address import Address
from .metadata import Metadata


class Transaction(object):
    """
    Represents a Cardano transaction.

    :param txid:            the ID of the transaction
    :param gross_amount:    gross amount, consisting of the amount of ADA paid and the fee
    :param fee:             fee amount in ADA
    :param inputs:          a sequence of :class:`Input` objects
    :param outputs:         a sequence of :class:`Output` objects
    :param metadata:        an instance of :class:`Metadata <cardano.metadata.Metadata>`
    """

    txid = None
    gross_amount = None
    fee = None
    inputs = None
    outputs = None
    inserted_at = None
    expires_at = None
    pending_since = None
    local_address = None
    metadata = None

    def __init__(self, txid=None, **kwargs):
        self.txid = txid or self.txid
        gross_amount = kwargs.pop("gross_amount", None)
        self.gross_amount = (
            gross_amount if gross_amount is not None else self.gross_amount
        )
        fee = kwargs.pop("fee", None)
        self.fee = fee if fee is not None else self.fee
        self.inputs = kwargs.pop("inputs", []) or (
            self.inputs if self.inputs is not None else []
        )
        self.outputs = kwargs.pop("outputs", []) or (
            self.outputs if self.outputs is not None else []
        )
        self.inserted_at = kwargs.pop("inserted_at", None) or self.inserted_at
        self.expires_at = kwargs.pop("expires_at", None) or self.expires_at
        self.pending_since = kwargs.pop("pending_since", None) or self.pending_since
        self.metadata = (
            kwargs.pop("metadata", Metadata()) or (self.metadata
            if self.metadata is not None
            else Metadata())
        )

    @property
    def amount(self):
        return self.gross_amount - self.fee


class IOBase(object):
    def __init__(self, address=None, amount=None):
        self.address = None if address is None else Address(address)
        self.amount = amount


class Input(IOBase):
    """
    Represents a :class:`Transaction` input.

    :param address: the origin address
    :type address:  :class:`cardano.address.Address`
    :param amount:  the amount in ADA
    :type amount:   :class:`Decimal`
    """

    def __init__(self, iid=None, address=None, amount=None):
        super(Input, self).__init__(address=address, amount=amount)
        self.id = iid


class Output(IOBase):
    """
    Represents a :class:`Transaction` output.

    :param address: the destination address
    :type address:  :class:`cardano.address.Address`
    :param amount:  the amount in ADA
    :type amount:   :class:`Decimal`
    """

    pass



from .address import Address
from . import exceptions


class WalletService(object):
    """
    Represents the service responsible for listing and retrieving the existing wallets or creating
    new ones.

    :param backend: the backend used to handle the underlying service layer

    """

    def __init__(self, backend=None):
        self.backend = backend

    def wallets(self):
        return [Wallet(wid, backend=self.backend) for wid in self.backend.wallet_ids()]

    def wallet(self, wid, passphrase=None):
        return Wallet(wid, backend=self.backend, passphrase=passphrase)

    def create_wallet(self, name, mnemonic, passphrase, mnemonic_2f=None):
        """
        Creates/restores a wallet. Returns only ID as the backend may need some time to sync
        before being able to return full wallet data.
        """
        return self.backend.create_wallet(name, mnemonic, passphrase, mnemonic_2f)


class Wallet(object):
    """
    Represents a single wallet. Allows for browsing the history, checking balance and spending
    funds.

    :param wid:         the wallet ID
    :param backend:     the backend used to handle the underlying service layer
    :param passphrase:  the passphrase protecting the wallet's spending functionality, not required
                        for read-only operations. It will be stored for the entire lifetime of the
                        object in ``.passphrase`` field. It might be also provided for each
                        individual spend operation, then it will be discarded after use.
    """

    passphrase = None

    def __init__(self, wid, backend, passphrase=None):
        self.wid = wid
        self.backend = backend
        self.passphrase = passphrase or self.passphrase
        if not self.backend.wallet_exists(wid):
            raise ValueError("Wallet of id '{:s}' doesn't exist.".format(wid))

    def sync_progress(self):
        return self.backend.sync_progress(self.wid)

    def addresses(self, with_usage=False):
        if with_usage:
            return [
                (Address(addr[0], wallet=self), addr[1])
                for addr in self.backend.addresses(self.wid)
            ]
        return [
            Address(addr[0], wallet=self) for addr in self.backend.addresses(self.wid)
        ]

    def balance(self):
        return self.backend.balance(self.wid)

    def assets(self):
        return self.backend.asset_balances(self.wid)

    def delete(self):
        return self.backend.delete_wallet(self.wid)

    def transactions(self):  # , start=None, end=None, order="ascending"):
        # WARNING: parameters don't really work; this is a known bug
        # if start:
        #    start = start.astimezone(tz=timezone.utc)
        # if end:
        #    end = end.astimezone(tz=timezone.utc)
        return self.backend.transactions(
            self.wid
        )  # , start=start, end=end, order=order)

    def _resolve_passphrase(self, passphrase):
        passphrase = passphrase or self.passphrase
        if passphrase is None:
            raise exceptions.MissingPassphrase(
                "Cannot perform operation without the passphrase"
            )
        return passphrase

    def transfer(self, address, amount, metadata=None, ttl=None, passphrase=None):
        """
        Sends a transfer from the wallet. Returns the resulting transaction.

        :param address: destination :class:`Address <cardano.address.Address>` or subtype
        :param amount: amount to send
        :param metadata: metadata to be sent, as :class:`Metadata <cardano.metadata.Metadata>`
                    instance od ``dict`` mapping ``int`` keys to values of acceptable types
        :param ttl: Time To Live in seconds. After TTL has lapsed the nodes give up on broadcasting
                    the transaction. Leave `None` to use the default value.
        :param passphrase: the passphrase to the wallet. It takes precedence over `self.passphrase`
                    and is discarded after use. If neither `self.passphrase` nor `passphrase` is
                    set, a :class:`MissingPassphrase <cardano.exceptions.MissingPassphrase>`
                    exception will be raised.
        :rtype: :class:`Transaction <cardano.transaction.Transaction>`
        """
        return self.backend.transfer(
            self.wid,
            ((address, amount),),
            metadata,
            ttl,
            self._resolve_passphrase(passphrase),
        )

    def transfer_multiple(self, destinations, metadata=None, ttl=None, passphrase=None):
        """
        Sends multiple transfers from the wallet. Returns the resulting transaction.

        :param destinations: a list of :class:`Address <cardano.address.Address>` and amount
                    pairs ``[(address, amount), ...]``
        :param metadata: metadata to be sent, as :class:`Metadata <cardano.metadata.Metadata>`
                    instance od ``dict`` mapping ``int`` keys to values of acceptable types
        :param ttl: Time To Live in seconds. After TTL has lapsed the nodes give up on broadcasting
                    the transaction. Leave `None` to use the default value.
        :param passphrase: the passphrase to the wallet. It takes precedence over `self.passphrase`
                    and is discarded after use. If neither `self.passphrase` nor `passphrase` is
                    set, a :class:`MissingPassphrase <cardano.exceptions.MissingPassphrase>`
                    exception will be raised.
        :rtype: :class:`Transaction <cardano.transaction.Transaction>`
        """
        return self.backend.transfer(
            self.wid,
            destinations,
            metadata,
            ttl,
            self._resolve_passphrase(passphrase),
        )

    def estimate_fee(self, destinations, metadata=None):
        """
        Estimates the fee for a potential transaction to specified destinations and carrying
        optional metadata. Returns a tuple of estimated minimum and maximum fee, in ADA.

        :param destinations: a list of :class:`Address <cardano.address.Address>` and amount
                    pairs ``[(address, amount), ...]``
        :param metadata: metadata to be sent, as :class:`Metadata <cardano.metadata.Metadata>`
                    instance od ``dict`` mapping ``int`` keys to values of acceptable types
        :rtype: (``Decimal``, ``Decimal``)
        """
        return self.backend.estimate_fee(self.wid, destinations, metadata)



from decimal import Decimal
import json
import logging
import operator
import requests

from ...metadata import Metadata
from ...numbers import from_lovelaces
from ...simpletypes import AssetID, Balance
from ...transaction import Transaction
from . import exceptions
from . import serializers

_log = logging.getLogger(__name__)


class JSONWithDecimalEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, Decimal):
            return str(Decimal)
        return super(JSONWithDecimalEncoder, self).default(o)


class WalletREST(object):
    base_url = None
    timeout = 10

    def __init__(self, protocol="http", host="localhost", port=8090):
        self.base_url = "{protocol}://{host}:{port}/v2/".format(
            protocol=protocol, host=host, port=port
        )
        _log.debug("WalletREST backend url: {:s}".format(self.base_url))

    def raw_request(self, method, path, params=None):
        url = "".join([self.base_url, path])
        hdr = {"Content-Type": "application/json"}
        params = params or {}
        _log.debug(
            u"{method} {url}\nParams:\n{params}".format(
                method=method,
                url=url,
                params=json.dumps(params, indent=2, sort_keys=True),
            )
        )
        rsp = getattr(requests, method.lower())(
            url, headers=hdr, data=json.dumps(params), timeout=self.timeout
        )
        if rsp.status_code != 204:  # if content exists
            result = rsp.json(parse_float=Decimal)
            _ppresult = json.dumps(
                result, cls=JSONWithDecimalEncoder, indent=2, sort_keys=True
            )
            _log.debug(u"Result:\n{result}".format(result=_ppresult))
        else:
            result = None
            _log.debug(u"No result (HTTP 204)")
        if rsp.status_code == 400:
            raise exceptions.BadRequest(result["message"], result=result)
        if rsp.status_code == 404:
            raise exceptions.NotFound(result["message"], result=result)
        return result

    def wallet_ids(self):
        return map(operator.itemgetter("id"), self.raw_request("GET", "wallets"))

    def wallet_exists(self, wid):
        try:
            self.raw_request("GET", "wallets/{:s}".format(wid))
        except exceptions.NotFound:
            return False
        return True

    def create_wallet(self, name, mnemonic, passphrase, mnemonic_2f=None):
        data = {
            "name": name,
            "mnemonic_sentence": mnemonic.split(),
            "passphrase": passphrase,
            "address_pool_gap": 20,
        }
        if mnemonic_2f:
            data["mnemonic_second_factor"] = mnemonic_2f.split()
        wdata = self.raw_request("POST", "wallets", data)
        return wdata["id"]

    def delete_wallet(self, wid):
        try:
            self.raw_request("DELETE", "wallets/{:s}".format(wid))
        except exceptions.NotFound:
            return False
        return True

    def sync_progress(self, wid):
        wdata = self.raw_request("GET", "wallets/{:s}".format(wid))
        return (
            1.0
            if wdata["state"]["status"] == "ready"
            else wdata["state"]["progress"]["quantity"] / Decimal(100)
        )

    def balance(self, wid):
        bdata = self.raw_request("GET", "wallets/{:s}".format(wid))["balance"]
        return Balance(
            from_lovelaces(bdata["total"]["quantity"]),
            from_lovelaces(bdata["available"]["quantity"]),
            from_lovelaces(bdata["reward"]["quantity"]),
        )

    def asset_balances(self, wid):
        data = self.raw_request("GET", "wallets/{:s}".format(wid))
        try:
            bdata = data["assets"]
        except KeyError:
            raise exceptions.NotSupported("Extra assets are not supported")
        assets = {}
        for ast in bdata["total"]:
            aid = AssetID(ast["asset_name"], ast["policy_id"])
            assets[aid] = assets[aid] if aid in assets else {}
            assets[aid]["total"] = ast["quantity"]
        for ast in bdata["available"]:
            aid = AssetID(ast["asset_name"], ast["policy_id"])
            assets[aid] = assets[aid] if aid in assets else {}
            assets[aid]["available"] = ast["quantity"]
        return {
            aid: Balance(bal["total"], bal["available"], None)
            for aid, bal in assets.items()
        }

    def addresses(self, wid):
        adata = self.raw_request("GET", "wallets/{:s}/addresses".format(wid))
        return [(ad["id"], True if ad["state"] == "used" else False) for ad in adata]

    def used_addresses(self, wid):
        return set(
            [ad[0] for ad in filter(operator.itemgetter(1), self.addresses(wid))]
        )

    def _txdata2tx(self, txd, used_addresses=None):
        inputs = (
            [serializers.get_input(inp) for inp in txd["inputs"]]
            if "inputs" in txd
            else []
        )
        outputs = (
            [serializers.get_output(outp) for outp in txd["outputs"]]
            if "outputs" in txd
            else []
        )
        local_inputs = set()
        local_outputs = set()
        if used_addresses:
            for out in outputs:
                if out.address and out.address in used_addresses:
                    local_outputs.add(out)
            for inp in inputs:
                if inp.address and inp.address in used_addresses:
                    local_inputs.add(inp)
        metadata = (
            Metadata.deserialize(txd["metadata"])
            if txd.get("metadata", None) is not None
            else None
        )
        return Transaction(
            txid=txd["id"],
            gross_amount=serializers.get_amount(txd["amount"]),
            fee=serializers.get_amount(txd["fee"]),
            inserted_at=serializers.get_block_position(txd["inserted_at"])
            if "inserted_at" in txd
            else None,
            expires_at=serializers.get_block_position(txd["expires_at"])
            if "expires_at" in txd
            else None,
            pending_since=serializers.get_block_position(txd["pending_since"])
            if "pending_since" in txd
            else None,
            inputs=inputs,
            outputs=outputs,
            metadata=metadata,
        )

    def transactions(self, wid, start=None, end=None, order="ascending"):
        data = {
            "order": order,
        }
        if start is not None:
            data["start"] = start.isoformat(timespec="seconds")
        if end is not None:
            data["end"] = end.isoformat(timespec="seconds")
        return [
            self._txdata2tx(txd, used_addresses=self.used_addresses(wid))
            for txd in self.raw_request(
                "GET", "wallets/{:s}/transactions".format(wid), data
            )
        ]

    def transfer(self, wid, destinations, metadata, ttl, passphrase):
        data = {
            "passphrase": passphrase,
            "payments": [
                {
                    "address": str(address),
                    "amount": serializers.store_amount(amount),
                }
                for (address, amount) in destinations
            ],
        }
        if metadata is not None:
            if not isinstance(metadata, Metadata):
                metadata = Metadata(metadata.items())
            data["metadata"] = metadata.serialize()
        if ttl is not None:
            data["time_to_live"] = serializers.store_interval(ttl)
        # NOTE: the order of the following two requests is important
        txd = self.raw_request("POST", "wallets/{:s}/transactions".format(wid), data)
        used_addresses = self.used_addresses(wid)
        return self._txdata2tx(txd, used_addresses)

    def estimate_fee(self, wid, destinations, metadata):
        data = {
            "payments": [
                {
                    "address": str(address),
                    "amount": serializers.store_amount(amount),
                }
                for (address, amount) in destinations
            ],
        }
        if metadata is not None:
            if not isinstance(metadata, Metadata):
                metadata = Metadata(metadata.items())
            data["metadata"] = metadata.serialize()
        feedata = self.raw_request(
            "POST", "wallets/{:s}/payment-fees".format(wid), data
        )
        return (
            serializers.get_amount(feedata["estimated_min"]),
            serializers.get_amount(feedata["estimated_max"]),
        )



import collections


Balance = collections.namedtuple("Balance", ["total", "available", "reward"])
Balance.__doc__ = "Represents a balance of asset, including total, principal and reward"
Balance.total.__doc__ = "The total balance"
Balance.available.__doc__ = "The principal, i.e. the total minus staking rewards"
Balance.reward.__doc__ = "The staking rewards (interest)"


BlockPosition = collections.namedtuple(
    "BlockPosition", ["epoch", "slot", "absolute_slot"]
)
BlockPosition.__doc__ = "Represents block's position within the blockchain"
BlockPosition.epoch.__doc__ = "Epoch number"
BlockPosition.slot.__doc__ = "Slot number"
BlockPosition.absolute_slot.__doc__ = "Absolute slot number"


class AssetID(object):
    """
    Represents the ID of a native Cardano asset. It consists of asset name and policy ID.
    It renders as string representation of ``policy_id:asset_name``.
    """

    asset_name = None
    policy_id = None

    def __init__(self, asset_name, policy_id):
        self.asset_name = asset_name if asset_name is not None else self.asset_name
        self.policy_id = policy_id or self.policy_id

    def __repr__(self):
        return "{:s}:{:s}".format(self.asset_name, self.policy_id)

    def __eq__(self, other):
        if isinstance(other, AssetID):
            return str(self) == str(other)
        elif isinstance(other, str):
            return str(self) == other
        elif isinstance(other, bytes):
            return str(self).encode() == other
        return super(AssetID, self).__eq__(other)

    def __hash__(self):
        return hash(str(self))



from ...address import Address
from ...numbers import from_lovelaces, to_lovelaces
from ...simpletypes import BlockPosition
from ...transaction import Input, Output


def get_amount(data):
    assert data["unit"] == "lovelace"
    return from_lovelaces(data["quantity"])


def store_amount(amount):
    return {"quantity": to_lovelaces(amount), "unit": "lovelace"}


def get_block_position(data):
    return BlockPosition(
        data["epoch_number"], data["slot_number"], data["absolute_slot_number"]
    )


def get_input(data):
    return Input(
        iid=data["id"],
        address=Address(data["address"]) if "address" in data else None,
        amount=get_amount(data["amount"]) if "amount" in data else None,
    )


def get_output(data):
    return Output(address=Address(data["address"]), amount=get_amount(data["amount"]))


def store_interval(seconds):
    return {"quantity": int(seconds), "unit": "second"}



from .consts import Era


class Address(object):
    """
    Represents Cardano address. Does no validation (TBD).
    Compares with ``str`` and ``bytes`` objects.

    :param addr:    the address as ``str`` or ``bytes`` or ``Address``
    """

    _address = ""
    era = None
    wallet = None

    def __init__(self, addr, wallet=None):
        addr = addr._address if isinstance(addr, Address) else addr
        self._address = addr
        self.wallet = wallet or self.wallet
        if addr.startswith("Ae2") or addr.startswith("DdzFF"):
            self.era = Era.BYRON
        elif addr.startswith("addr1") or addr.startswith("addr_test1"):
            self.era = Era.SHELLEY
        else:
            raise ValueError("String {} is not a valid Cardano address")

    def __repr__(self):
        return self._address

    def __eq__(self, other):
        if isinstance(other, Address):
            return str(self) == str(other)
        elif isinstance(other, str):
            return str(self) == other
        elif isinstance(other, bytes):
            return str(self).encode() == other
        return super(Address, self).__eq__(other)

    def __hash__(self):
        return hash(str(self))

    def __format__(self, spec):
        return format(str(self), spec)



from decimal import Decimal
import operator

_MININT = -(2 ** 64 - 1)
_MAXINT = 2 ** 64 - 1


class Metadata(dict):
    TYPE_RESOLVERS = {
        "string": lambda s: s,
        "int": lambda i: Metadata.deserialize_int(i),
        "bytes": bytes.fromhex,
        "list": lambda l: [
            Metadata.TYPE_RESOLVERS[k](v)
            for k, v in map(
                operator.itemgetter(0),
                map(list, map(operator.methodcaller("items"), l)),
            )
        ],
        "map": lambda m: Metadata.deserialize_map(m),
    }

    @staticmethod
    def validate_key(key):
        if not isinstance(key, int):
            raise KeyError("Metadata keys must be integers")
        if key < 0:
            raise KeyError("Metadata key {:d} is negative".format(key))
        if key > _MAXINT:
            raise KeyError("Metadata key {:d} is over 2^64-1".format(key))
        return key

    @staticmethod
    def validate_value(val):
        if isinstance(val, (str, bytes, bytearray)):
            if len(val) > 64:
                raise ValueError(
                    "The string {} is too long ({:d} > 64)".format(val, len(val))
                )
        elif isinstance(val, int):
            if val < _MININT:
                raise ValueError("Int {:d} is less than -2^64-1".format(val))
            elif val > _MAXINT:
                raise ValueError("Int {:d} is over 2^64-1".format(val))
        elif isinstance(val, list):
            pass
        elif isinstance(val, dict):
            for k, v in val.items():
                Metadata.validate_value(k)
                Metadata.validate_value(v)
        else:
            raise TypeError(
                "Metadata values must be of int, str, bytes, bytearray, lists of thereof or another Metadata instances, not {}".format(
                    str(type(val))
                )
            )
        return val

    @staticmethod
    def serialize_value(val):
        if isinstance(val, int):
            return {"int": val}
        elif isinstance(val, str):
            return {"string": val}
        elif isinstance(val, (bytes, bytearray)):
            return {"bytes": val.hex()}
        elif isinstance(val, list):
            return {"list": [Metadata.serialize_value(i) for i in val]}
        elif isinstance(val, dict):
            return {
                "map": [
                    {"k": Metadata.serialize_value(k), "v": Metadata.serialize_value(v)}
                    for k, v in val.items()
                ]
            }
        # This should never happen
        raise RuntimeError(
            "Found unserializable value of {} (type {})".format(val, str(type(val)))
        )

    @staticmethod
    def deserialize_item(key, vdata):
        if len(vdata) > 1:
            raise ValueError(
                "The value dict for key {:s} has {:d} members while only one is permitted".format(
                    key, len(vdata)
                )
            )
        typename = list(vdata.keys()).pop()
        val = list(vdata.values()).pop()
        return {int(key): Metadata.TYPE_RESOLVERS[typename](val)}

    @staticmethod
    def deserialize_int(i):
        if isinstance(i, int):
            return i
        elif isinstance(i, Decimal):
            return int(i.quantize(1))
        raise TypeError("Got int serialized as {} of value {}".format(str(type(i)), i))

    @staticmethod
    def deserialize_map(themap):
        data = {}
        for m in themap:
            ktype, kval = list(m["k"].items()).pop()
            vtype, vval = list(m["v"].items()).pop()
            data[Metadata.TYPE_RESOLVERS[ktype](kval)] = Metadata.TYPE_RESOLVERS[vtype](
                vval
            )
        return data

    @staticmethod
    def deserialize(txdata):
        """
        Deserializes transaction metadata ``dict`` and returns :class:``Metadata`` instance.
        :param txdata:  the transaction data
        """
        data = {}
        for key, vdata in txdata.items():
            data.update(Metadata.deserialize_item(key, vdata))
        return Metadata(data.items())

    def __init__(self, *args, **kwargs):
        if len(args) > 0:
            for k, v in args[0]:
                Metadata.validate_key(k)
                Metadata.validate_value(v)
        super(Metadata, self).__init__(*args, **kwargs)

    def __setitem__(self, key, val):
        return super(Metadata, self).__setitem__(
            Metadata.validate_key(key), Metadata.validate_value(val)
        )

    def serialize(self):
        return {str(k): Metadata.serialize_value(v) for k, v in self.items()}



class CardanoException(Exception):
    pass


class BackendException(CardanoException):
    pass


class WalletException(CardanoException):
    pass


class MissingPassphrase(WalletException):
    pass



import enum


class Era(enum.IntEnum):
    """
    Represents Cardano era, a distinct phase of platform development.

    .. warning:: **Do NOT use the integer values directly.**
                There are new Eras being introduced, as for example Allegra and Mary were
                inserted as stepping stones into full Goguen. However, you may use comparison
                operators between them, to check which was earlier or later than the other one.
    """

    BYRON = 1
    SHELLEY = 2
    ALLEGRA = 3
    MARY = 4
    GOGUEN = 5
    BASHO = 6
    VOLTAIRE = 7
